/*
 *
 * Copyrighted by NPAC, Syracuse Univ., 1994.
 *
 */

#include "DEM.h"
#include <errno.h>
#include <ctype.h>

DEM *read_DEM_file(char *DataFilename)
{
   char        *Subroutine_Name = "read_DEM_file()";
   FILE        *DataFile;
   DEM         *DEM_Data;
   int          N;
   int          Max_M;
   DEM_record **Records;
   int          i;

   if ((DataFile = fopen(DataFilename, "r")) == NULL)
   {
      printf("%s(): cannot open %s\n", Subroutine_Name, DataFilename);
      exit(1);
   }

   if ((DEM_Data = (DEM *) malloc(sizeof(DEM))) == NULL)
   {
      perror(Subroutine_Name);
      exit(1);
   }

   DEM_Data -> Header = read_DEM_header(DataFile);

   N = DEM_Data -> Header -> N;
#if 0
   if ((Matrix = (DEM_matrix *) malloc(sizeof(DEM_matrix))) == NULL)
   {
      
   }
#endif
   if ((Records = (DEM_record **) malloc(N * sizeof(DEM_record *))) == NULL)
   {
      perror(Subroutine_Name);
      exit(1);
   }

   DEM_Data -> records = Records;

   Max_M = -1;
   for (i = 0; i < N; i++)
   {
      Records[i] = read_DEM_record(DataFile);
      if (Max_M < Records[i] -> M)
         Max_M = Records[i] -> M;
   }

   DEM_Data -> Header -> M = Max_M;

#if 0
   if (DEM_Data -> Header.Accuracy_Code != 0)
      DEM_Data -> Stat = read_DEM_stat(DataFile);
   else
      DEM_Data -> Stat = NULL;
#endif

   fclose(DataFile);
   return(DEM_Data);
}

DEM_header *read_DEM_header(FILE *DataFile)
{
   char       *Subroutine_Name = "read_DEM_header()";
   DEM_header *Header;
   int         i;
   int         j;

   if ((Header = (DEM_header *) malloc(sizeof(DEM_header))) == NULL)
   {
      perror(Subroutine_Name);
      exit(1);
   }

   for (i = 0; i < 144; i++)
     Header -> Name[i] = (char) fgetc(DataFile);

   Header -> Level_code = read_int(DataFile, 6);

   Header -> Pattern_code = read_int(DataFile, 6);

   Header -> Planimetric_code = read_int(DataFile, 6);

   Header -> Zone_code = read_int(DataFile, 6);

   for (i = 0; i < 15; i++)
      Header -> Map_Projection_Parameters[i] = read_double(DataFile, 24);

   Header -> Ground_Planimetric_Coordinates_Units_Code = read_int(DataFile, 6);

   Header -> Elevation_Coordinates_Units_Code = read_int(DataFile, 6);

   Header -> Polygon_Sides = read_int(DataFile, 6);

   for (i = 0; i < 4; i++)
      for (j = 0; j < 2; j++)
         Header -> Ground_Coordinates[i][j] = read_double(DataFile, 24);

   Header -> Min_Elevation = read_double(DataFile, 24);

   Header -> Max_Elevation = read_double(DataFile, 24);

   Header -> Rotation_Angle = read_double(DataFile, 24);

   Header -> Accuracy_Code = read_int(DataFile, 6);

   for (i = 0; i < 3; i++)
      Header -> Spatial_Resolution[i] = read_double(DataFile, 12);

   Header -> Row_per_record = read_int(DataFile, 6);

   Header -> N = read_int(DataFile, 6);

   for (i = 0; i < 160; i++)
      fgetc(DataFile);

   return(Header);
}

DEM_record *read_DEM_record(FILE *DataFile)
{
   char       *Subroutine_Name = "read_DEM_record()";
   DEM_record *Record;
   int         M;
   int        *Elevations;
   int         Index;
   int         i;
   int         limit;
   int         padding;

   if ((Record = (DEM_record *) malloc(sizeof(DEM_record))) == NULL)
   {
      perror(Subroutine_Name);
      exit(1);
   }

   Record -> m = read_int(DataFile, 6);
   Record -> n = read_int(DataFile, 6);
   Record -> M = M = read_int(DataFile, 6);
   Record -> Col_per_record = read_int(DataFile, 6);

   if ((Elevations = (int *) malloc(M * sizeof(int))) == NULL)
   {
      perror(Subroutine_Name);
      exit(1);
   }

   Record -> Elevations = Elevations;
   Record -> Coordinate[0] = read_double(DataFile, 24);
   Record -> Coordinate[1] = read_double(DataFile, 24);

   Record -> Base_Elevation = read_double(DataFile, 24);

   Record -> Min_Elevation = read_double(DataFile, 24);
   Record -> Max_Elevation = read_double(DataFile, 24);

   padding = USGS_FILE_BLOCK_SIZE
             - (TYPE_B_HEADER_NUM_INT * USGS_INT_FIELD_WIDTH)
             - (TYPE_B_HEADER_NUM_DOBULE * USGS_DOUBLE_FIELD_WIDTH);

   limit = padding / USGS_INT_FIELD_WIDTH;
   padding = padding % USGS_INT_FIELD_WIDTH;

   Index = 0;
   M -= limit;

   for (i = 0; i < limit; i++)
   {
      Elevations[Index] = read_int(DataFile, 6);
      Index++;
   }

   for (i = 0; i < padding; i++)
         fgetc(DataFile);

   limit = USGS_FILE_BLOCK_SIZE / USGS_INT_FIELD_WIDTH;
   padding = USGS_FILE_BLOCK_SIZE % USGS_INT_FIELD_WIDTH;

   while (M >= limit)
   {
      for (i = 0; i < limit; i++)
      {
         Elevations[Index] = read_int(DataFile, 6);
         Index++;
      }

      for (i = 0; i < padding; i++)
         fgetc(DataFile);
      M -= limit;
   }

   for (i = 0; i < M; i++)
   {
      Elevations[Index] = read_int(DataFile, 6);
      Index++;
   }

   padding = USGS_FILE_BLOCK_SIZE - (M * USGS_INT_FIELD_WIDTH);

   for (i = 0; i < padding; i++)
      fgetc(DataFile);

   return(Record);
}

#if 0
DEM_stat *read_DEM_stat(FILE *DataFile)
{
   char       *Subroutine_Name = "read_DEM_record()";
   DEM_accuracy *Record;
}
#endif

#if 0
#define DEBUG_read_int
#define DEBUG_read_double
#endif

int read_int(FILE *DataFile, int field_width)
{
   char Chr;
   char Buffer[128];
   int  Integer;
   int  i;

   for (i = 0; i < field_width; i++)
      Buffer[i] = fgetc(DataFile);
   Buffer[i] = '\0';

#ifdef DEBUG_read_int
   printf("read_int(): Buffer = (%s)\n", Buffer);
#endif

   sscanf(Buffer, "%d", &Integer);

   return(Integer);
}

double read_double(FILE *DataFile, int field_width)
{
   char   Chr;
   char   Buffer[128];
   double Double;
   int    i;
   int    j;

   for (i = 0; i < field_width; i++)
      if ((Chr = fgetc(DataFile)) != 'D')
         Buffer[i] = Chr;
      else
         Buffer[i] = 'E';
   Buffer[i] = '\0';

#ifdef DEBUG_read_double
   printf("read_double(): Buffer = (%s)\n", Buffer);
#endif

   sscanf(Buffer, "%lf", &Double);

   return(Double);
}

void print_DEM_header(DEM_header *Header)
{
   int i;
   int j;

   printf("DEM header:\n");
   printf("Name = %s\n", Header -> Name);
   printf("Level_code = %d\n", Header -> Level_code);
   printf("Pattern_code = %d\n", Header -> Pattern_code);
   printf("Planimetric_code = %d\n", Header -> Planimetric_code);
   printf("Zone_code = %d\n", Header -> Zone_code);
   for (i = 0; i < 15; i++)
      printf("Map_Projection_Parameters[%d] = %g\n", i,
             Header -> Map_Projection_Parameters[i]);

   printf("Ground_Planimetric_Coordinates_Units_Code = %d\n",
          Header -> Ground_Planimetric_Coordinates_Units_Code);

   printf("Elevation_Coordinates_Units_Code = %d\n",
          Header -> Elevation_Coordinates_Units_Code);

   printf("Polygon_Sides = %d\n", Header -> Polygon_Sides);

   for (i = 0; i < 4; i++)
      for (j = 0; j < 2; j++)
         printf("Ground_Coordinates[%d][%d] = %g\n", i, j,
                Header -> Ground_Coordinates[i][j] / 3600.0);

   printf("Min_Elevation = %g\n", Header -> Min_Elevation);

   printf("Max_Elevation = %g\n", Header -> Max_Elevation);

   printf("Rotation_Angle = %g\n", Header -> Rotation_Angle);

   printf("Accuracy_Code = %d\n", Header -> Accuracy_Code);

   for (i = 0; i < 3; i++)
      printf("Spatial_Resolution[%d] = %g\n", i,
             Header -> Spatial_Resolution[i]);

   printf("Row_per_record = %d\n", Header -> Row_per_record);

   printf("N = %d\n", Header -> N);
   printf("M = %d\n", Header -> M);
}

void print_DEM_record(DEM_record *Record)
{
   int j;
   int M;
   int *Elevations;

   printf("(m, n) = (%d, %d)\n", Record -> m, Record -> n);
   M = Record -> M;
   printf("(M, N) = (%d, %d)\n", Record -> M,
          Record -> Col_per_record);
   printf("Ground_Coordinates = (%g, %g)\n",
          (Record -> Coordinate)[0] / 3600.0,
          (Record -> Coordinate)[1] / 3600.0);
   printf("Base_Elevation = %g\n", Record -> Base_Elevation);
   printf("Max_Elevation = %g\n", Record -> Max_Elevation);
   printf("Min_Elevation = %g\n", Record -> Min_Elevation);

   Elevations = Record -> Elevations;
   for (j = 0; j < M; j++)
      printf("Elevations[%d] = %d\n", j, Elevations[j]);
}

void print_DEM(DEM *DEM_data)
{
   DEM_header  *Header  = DEM_data -> Header;
   DEM_record **Records = DEM_data -> records;
   int i;
   int M;
   int N;

   print_DEM_header(Header);

   N = Header -> N;

   for (i = 0; i < N; i++)
      print_DEM_record(Records[i]);
}
